import subprocess


def listen_ports():
    output = subprocess.check_output(["netstat", "-lnt"]).decode()
    ports = set()
    for line in output.split("\n"):
        if "LISTEN" in line:
            fields = line.split()
            field = fields[3]
            addr, _, port = field.partition(":")
            ports.add(int(port))
    return ports



import os

from rstms_testmail import gmail

TEST_FROM = os.environ["TESTMAIL_FROM"]
TEST_TO = os.environ["TESTMAIL_TO"]


def disable_test_gmail_init():
    server = gmail.Gmail()
    assert server


def disable_test_gmail_send():
    server = gmail.Gmail()
    ret = server.send(TEST_FROM, TEST_TO, "testmail_test", "testmail_message")
    assert ret



import base64
import os
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from .watcher import PortWatcher

# If modifying these SCOPES, delete the file token.json.
SCOPES = ["https://www.googleapis.com/auth/gmail.send"]


class Gmail:

    def __init__(self):
        with PortWatcher("ssh -q -N -R {}:localhost:{} beaker"):
            self.creds = self.authenticate_gmail_api()
            self.service = build("gmail", "v1", credentials=self.creds)

    def authenticate_gmail_api(self):
        """Shows basic usage of the Gmail API."""

        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first time.
        if os.path.exists("token.json"):
            creds = Credentials.from_authorized_user_file("token.json", SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file("credentials.json", SCOPES)
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open("token.json", "w") as token:
                token.write(creds.to_json())

        return creds

    def send(self, from_addr, to_addr, subject, message_text):
        """Create and send an email message."""
        service = self.service
        message = MIMEMultipart()
        message["to"] = to_addr
        message["from"] = from_addr
        message["subject"] = subject
        msg = MIMEText(message_text)
        message.attach(msg)

        # Encode the message in base64
        raw = base64.urlsafe_b64encode(message.as_bytes())
        raw = raw.decode()
        body = {"raw": raw}

        try:
            message = service.users().messages().send(userId="me", body=body).execute()
            print("Message Id: %s" % message["id"])
            return message
        except Exception as error:
            print(f"An error occurred: {error}")



"""Console script for rstms_testmail."""

import subprocess
import sys
from pprint import pformat
import json

import click
import click.core

from .counter import Counter
from .settings import Settings
from .exception_handler import ExceptionHandler
from .gmail import Gmail
from .sendgrid_server import SendGrid
from .smtp_server import SMTPServer
from .shell import _shell_completion
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def _ehandler(ctx, option, debug):
    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug
    return debug


def fail(msg):
    click.echo("Failed: " + msg, err=True)
    sys.exit(-1)


@click.command("testmail")
@click.version_option(message=header)
@click.option("-d", "--debug", is_eager=True, is_flag=True, callback=_ehandler, help="debug mode")
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.option('-q', '--quiet', is_flag=True)
@click.option('-v', '--verbose', is_flag=True)
@click.option("-f", "--from", "from_addr", envvar="TESTMAIL_FROM")
@click.option("-t", "--to", "to_addr", envvar="TESTMAIL_TO")
@click.option("-s", "--subject", default="test{}")
@click.option("-c", "--set-counter", type=int)
@click.option("-s", "--system", envvar="TESTMAIL_SYSTEM")
@click.option("-e", "--exec", "exec_command", envvar="TESTMAIL_EXEC")
@click.option("-k", "--api-key", envvar="TESTMAIL_API_KEY")
@click.option("-u", "--username", envvar="TESTMAIL_USERNAME")
@click.option("-p", "--password", envvar="TESTMAIL_PASSWORD")
@click.option("-P", "--port", default=465, envvar="TESTMAIL_PORT")
@click.option('--profile', default='default', envvar='TESTMAIL_PROFILE')
@click.option("--dryrun", is_flag=True)
@click.argument("message", required=False)
@click.pass_context
def cli(ctx, debug, shell_completion, quiet, verbose, dryrun, to_addr, from_addr, set_counter, subject, message, system, exec_command, username, password, port, api_key, profile):
    """send a test email"""

    if "{}" in subject:
        counter = Counter(label="testmail")
        count = counter.bump(set_counter)
        subject = subject.replace("{}", str(count))

    settings = Settings(profile=f"testmail_{profile}", from_addr=from_addr, to_addr=to_addr, system=system, exec_command=exec_command, username=username, password=password,port=port, api_key=api_key)

    if debug:
        click.echo(json.dumps({'config': settings.dict()}, indent=2))

    if settings.exec_command is not None:
        message = subprocess.check_output(settings.exec_command, shell=True).decode()

    if message == "-":
        message = sys.stdin.read()
    elif message is None:
        message = subject

    if settings.system == "gmail":
        server = Gmail()
    elif settings.system == "sendgrid":
        server = SendGrid(api_key)
    elif settings.system.startswith('smtp:'):
        server = SMTPServer(settings.system, settings.port, settings.username, settings.password)
    else:
        fail("unknown system")

    if dryrun:
        click.echo(pformat(settings.dict()))
        sys.exit(0)
    else:
        error = server.send(settings.from_addr, settings.to_addr, subject, message)

    if error:
        fail(pformat(error))

    if not quiet:
        click.echo(f"sent {subject} to {settings.to_addr}")

    if verbose:
        click.echo(pformat(server.result))

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



from rstms_testmail.counter import Counter


def test_counter():
    counter = Counter(dir=".", count=0)
    assert counter.value == 0

    counter.bump()
    assert counter.value == 1

    ret = counter.bump(10)
    assert ret == 10
    assert counter.value == 10

    counter2 = Counter(dir=".")
    assert counter2.value == 10



from pathlib import Path
import json

class Settings:
    def __init__(self, **kwargs):
        profile=kwargs.pop('profile', "")
        dir = kwargs.pop('dir', '~')
        if dir in ['~', None]:
            dir = Path.home()
        else:
            dir = Path(dir)
        self.file = dir / f".{profile}_settings"

        old_settings = self.read()

        self._settings = {}

        for k,v in kwargs.items():
            if v is None:
                v = old_settings.get(k, None)
            self._settings[k] = v
            setattr(self, k, v)

        self.write(self._settings)

    def read(self):
        if self.file.is_file():
            content = self.file.read_text()
        else:
            content = "{}"
        return json.loads(content)

    def write(self, settings):
        self.file.write_text(json.dumps(settings))

    def dict(self):
        return self._settings



# persistent counter

from pathlib import Path


class Counter:
    def __init__(self, dir=None, label=None, count=None):

        if dir in ["~", None]:
            dir = Path.home()
        else:
            dir = Path(dir)

        if label is None:
            filename = ".counter"
        else:
            filename = f".{label}_counter"

        self.file = dir / filename

        if count is None:
            self.read()
        else:
            self.value = int(count)
            self.write()

    def read(self):
        if not self.file.is_file():
            self.value = 0
            return self.write()
        self.value = int(self.file.read_text() or "0")
        return self.value

    def write(self):
        self.file.write_text(str(self.value) + "\n")
        return self.value

    def bump(self, value=None):
        if value:
            self.value = value
        else:
            self.value += 1

        return self.write()



import os

from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail


def test_sendgrid_send():
    message = Mail(
        from_email=os.environ["TESTMAIL_FROM"],
        to_emails=os.environ["TESTMAIL_TO"],
        subject="Sending with Twilio SendGrid is Fun",
        plain_text_content="this is plain text email",
    )

    sendgrid_client = SendGridAPIClient(os.environ.get("SENDGRID_API_KEY"))
    response = sendgrid_client.send(message)
    print(response.status_code)
    print(response.body)
    print(response.headers)



import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

class SMTPServer:

    def __init__(self, system, port, username, password):
        _, _, self.server = system.partition(':')
        self.port = int(port)
        self.username = username
        self.password = password

    def send(self, from_addr, to_addr, subject, message):
        """
        Send an email using an SMTP server that requires SSL and authentication.

        Parameters:
        - from_addr: str - the sender's email address
        - to_addr: str - the recipient's email address
        - subject: str - the subject of the email
        - message: str - the body of the email
        """
        # Create the container email message.
        mime_message = MIMEMultipart()
        mime_message['Subject'] = subject
        mime_message['From'] = from_addr
        mime_message['To'] = to_addr
        mime_message.attach(MIMEText(message, 'plain'))

        # Connect to the SMTP server using SSL
        server = smtplib.SMTP_SSL(self.server, self.port)
        server.login(self.username, self.password)
        server.sendmail(from_addr, to_addr, mime_message.as_string())
        server.quit()
        return None



from os import system
from subprocess import Popen
from time import sleep

import pytest

from rstms_testmail.netstat import listen_ports
from rstms_testmail.watcher import PortWatcher


@pytest.fixture
def context():
    system("pkill nc")
    yield None
    system("pkill nc")
    system("pgrep -a ssh | grep 10000")


def test_watcher_listen(context):
    with PortWatcher("ssh -q -N -R {}:localhost:{} beaker") as watcher:
        assert watcher.running()
        assert not watcher.proc
        assert 10000 not in listen_ports()
        netcat = Popen(["nc", "-lp", "10000"])
        sleep(1)
        assert 10000 in listen_ports()
        sleep(2)
        assert watcher.proc
        netcat.terminate()
        netcat.wait()
        assert 10000 not in listen_ports()
        while watcher.running() is True:
            print("tick")
            sleep(1)
    assert not watcher.running()
    assert 10000 not in listen_ports()


def test_watcher_none(context):
    with PortWatcher("ssh -q -N -R {}:localhost:{} beaker") as watcher:
        assert watcher.running()
        assert not watcher.proc
        sleep(1)
    assert not watcher.running()
